/* Author: Minjae Wee	mwee@uci.edu */
// Wumpus World - Average AI.cpp

#include "AverageAgent.hpp"

AverageAgent::AverageAgent()
	:Agent()
{
	move_num = 0;
	row = 25;
	col = 25;
	for (int i = 0; i < row; i++){
		std::vector<Cell> temp;
		for (int j = 0; j < col; j++){
			Cell c;
			c.x = i;
			c.y = j;
			c.pit = 0;
			c.wumpus = 0;
			c.stench = -1;
			c.breeze = -1;
			c.visited = 0;
			temp.push_back(c);
		}
		board.push_back(temp);
	}
	board[0][0].stench = 0;
	board[0][0].breeze = 0;
	board[0][0].visited = 1;
	position = std::pair<int, int>(0, 0);	// start @ cell (0,0)
	facing = std::pair<int, int>(0, 1);		// start facing right

	gold = false;
	gotGold = false;
	notShot = true;
	wumpusKilled = false;

	lastAction = NIL;
}

// Returns true if a given percept is @ current position; otherwise false
bool AverageAgent::checkPerceptIn(Percept p){
	const bool is_in = ps.find(p) != ps.end();
	return is_in;
}

void AverageAgent::visitCell(){
	for (int i = 0; i < row; i++){
		for (int j = 0; j < col; j++){
			board[i][j].wumpus = 0;
			board[i][j].pit = 0;
		}
	}

	for (int i = 0; i < row; i++){
		for (int j = 0; j < col; j++){
			if (board[i][j].stench == 1){
				//std::cout << "updated wumpus likelihood" << std::endl;
				updateWumpusPitProb(i, j, 'W', 1, false);
			}
			if (board[i][j].breeze == 1){
				//std::cout << "updated pit likelihood" << std::endl;
				updateWumpusPitProb(i, j, 'P', 1, false);
			}
		}
	}

	for (int i = 0; i < row; i++){
		for (int j = 0; j < col; j++){
			// Visited already => safe (no pit, no wumpus)
			if (board[i][j].visited){
				board[i][j].wumpus = 0;
				board[i][j].pit = 0;
				if (board[i][j].stench == 0)
					updateWumpusPitProb(i, j, 'W', 0, true);
				if (board[i][j].breeze == 0)
					updateWumpusPitProb(i, j, 'P', 0, true);
			}
		}
	}

	for (int i = 0; i < row; i++){
		for (int j = 0; j < col; j++){
			if (board[i][j].visited){
				board[i][j].wumpus = 0;
				board[i][j].pit = 0;
			}
		}
	}
}

void AverageAgent::updateWumpusPitProb(int r, int c, char wum_or_pit, int prob, bool setvalue){
	std::vector<std::pair<int, int>> temp_adj_cells;
	temp_adj_cells.push_back(std::pair<int, int>(r, c - 1));
	temp_adj_cells.push_back(std::pair<int, int>(r, c + 1));
	temp_adj_cells.push_back(std::pair<int, int>(r - 1, c));
	temp_adj_cells.push_back(std::pair<int, int>(r + 1, c));

	// Check if adjacent cell coordinates are on board
	std::vector<std::pair<int, int>> adj_cells;
	for (int i = 0; i < temp_adj_cells.size(); i++){
		int a = temp_adj_cells[i].first;
		int b = temp_adj_cells[i].second;
		if (a >= 0 && a < row && b >= 0 && b < col){
			adj_cells.push_back(std::pair<int, int>(a, b));
		}
	}
		
	for (int i = 0; i < adj_cells.size(); i++){
		int a = adj_cells[i].first;
		int b = adj_cells[i].second;
		if (wum_or_pit == 'W'){
			if (!setvalue)
				board[a][b].wumpus += prob;
			else
				board[a][b].wumpus = prob;
		}
		else if (wum_or_pit == 'P'){
			if (!setvalue)
				board[a][b].pit += prob;
			else
				board[a][b].pit = prob;
		}
	}
}

void AverageAgent::placeExaminePercepts(){
	if (checkPerceptIn(Agent::STENCH)){
		board[position.first][position.second].stench = 1;
	}
	else{
		board[position.first][position.second].stench = 0;
	}
	if (checkPerceptIn(Agent::BREEZE)){
		board[position.first][position.second].breeze = 1;
	}
	else{
		board[position.first][position.second].breeze = 0;
	}
	visitCell();
	if (move_proc.empty()){
		board[position.first][position.second].visited++;
	}
}

void AverageAgent::updateBoardSize(){
	// if agent facing right
	if (facing.first == 0 && facing.second == 1){
		col = position.second + 1;
		//std::cout << "resized/updated column to: " << col << std::endl;
		for (int j = 0; j < board.size(); j++)
		{
			board[j].resize(col);
		}
	}
	// if agent facing down
	if (facing.first == 1 && facing.second == 0){
		row = position.first + 1;
		//std::cout << "resized/updated row to: " << row << std::endl;
		board.resize(row);
	}
}

void AverageAgent::planMove(std::pair<std::string, Cell> move){
	//std::cout << "\t\tPlanning to move: " << move.first << move.second.x << move.second.y << std::endl;
	if (move.first == "climb"){
		move_proc.clear();
		move_proc.push_back(CLIMB);
		return;
	}

	if (move.first == "pass"){
		move_proc.clear();
		move_proc.push_back(TURN_RIGHT);
		return;
	}

	std::string goal = facingToString(facing);
	if (move.second.x < position.first){
		goal = "up";
	}
	if (move.second.x > position.first){
		goal = "down";
	}
	if (move.second.y < position.second){
		goal = "left";
	}
	if (move.second.y > position.second){
		goal = "right";
	}

	std::vector<Action> temp_proc1;
	std::vector<Action> temp_proc2;
	std::string temp_facing;

	temp_facing = facingToString(facing);
	while (temp_facing != goal){
		//std::cout << "\t\t\tGoal = " << goal << "\t\tTempFacing = " << temp_facing << std::endl;
		temp_proc1.push_back(TURN_RIGHT);
		temp_facing = turnRight(temp_facing);
	}

	temp_facing = facingToString(facing);
	while (temp_facing != goal){
		temp_proc2.push_back(TURN_LEFT);
		temp_facing = turnLeft(temp_facing);
	}

	if (temp_proc1.size() < temp_proc2.size())
		move_proc = temp_proc1;
	else
		move_proc = temp_proc2;

	if (move.first == "shoot")
		move_proc.push_back(SHOOT);
	else if (move.first == "climb")
		move_proc.push_back(CLIMB);
	else
		move_proc.push_back(FORWARD);
}


std::pair<std::string, Cell> AverageAgent::decideMove(){
	Cell none;
	if (gold && position.first == 0 && position.second == 0){
		return std::pair<std::string, Cell>("climb", none);
	}

	std::vector<Cell> possible_moves = getPossibleMoves();
	std::sort(possible_moves.begin(), possible_moves.end(), sort_danger());

	for (int i = 0; i < possible_moves.size(); i++){
		if (possible_moves[i].wumpus >= 2 && !wumpusKilled && notShot)
			return std::pair<std::string, Cell>("shoot", possible_moves[i]);
	}

	for (int i = 0; i < possible_moves.size(); i++){
		if (!possible_moves[i].visited && !possible_moves[i].wumpus && !possible_moves[i].pit)
			return std::pair<std::string, Cell>("move", possible_moves[i]);
	}

	for (int i = 0; i < possible_moves.size(); i++){
		if (possible_moves[i].visited <= 1 && possible_moves[i].wumpus <= 1 && possible_moves[i].pit == 0){
			if (possible_moves[i].wumpus != 0 && notShot){
				return std::pair<std::string, Cell>("shoot", possible_moves[i]);
			}
			return std::pair<std::string, Cell>("move", possible_moves[i]);
		}
	}

	for (int i = 0; i < possible_moves.size(); i++){
		if (possible_moves[i].visited <= 2 && possible_moves[i].wumpus <= 2 && possible_moves[i].pit == 0){
			if (possible_moves[i].wumpus != 0 && notShot){
				return std::pair<std::string, Cell>("shoot", possible_moves[i]);
			}
			return std::pair<std::string, Cell>("move", possible_moves[i]);
		}
	}

	for (int i = 0; i < possible_moves.size(); i++){
		if (possible_moves[i].visited <= 3 && possible_moves[i].wumpus <= 3 && possible_moves[i].pit == 0){
			if (possible_moves[i].wumpus != 0 && notShot){
				return std::pair<std::string, Cell>("shoot", possible_moves[i]);
			}
			return std::pair<std::string, Cell>("move", possible_moves[i]);
		}
	}
	// std::cout << "No safe move to make. Start backtracking!" << std::endl;
	gold = true;
	return std::pair<std::string, Cell>("pass", none);
}

// Return all 2 ~ 4 adjacent cell positions, given the current cell position
std::vector<Cell> AverageAgent::getPossibleMoves(){
	std::vector<Cell> cells;
	int x = position.first;
	int y = position.second;
	if (x >= 0 && x < row && y - 1 >= 0 && y - 1 < col){ // left cell
		cells.push_back(board[x][y - 1]);
	}
	if (x >= 0 && x < row && y + 1 >= 0 && y + 1 < col){ // right cell 
		cells.push_back(board[x][y + 1]);
	}
	if (x - 1 >= 0 && x - 1 < row && y >= 0 && y < col){ // top cell 
		cells.push_back(board[x - 1][y]);
	}
	if (x + 1 >= 0 && x + 1 < row && y >= 0 && y < col){ // bottom cell
		cells.push_back(board[x + 1][y]);

	}
	/*
	for (int i = 0; i < cells.size(); i++){
		std::cout << "Possible move" << i << ": " << cells[i].x << cells[i].y << "\n\n";
	}
	*/
	return cells;
}


std::string AverageAgent::facingToString(std::pair<int, int> facing){
	std::string result;
	if (facing.first == 1 && facing.second == 0)
		result = "down";
	else if (facing.first == -1 && facing.second == 0)
		result = "up";
	else if (facing.first == 0 && facing.second == 1)
		result = "right";
	else if (facing.first == 0 && facing.second == -1)
		result = "left";
	return result;
}

std::string AverageAgent::turnRight(std::string facing_temp){
	if (facing_temp == "down")
		facing_temp = "left";
	else if (facing_temp == "left")
		facing_temp = "up";
	else if (facing_temp == "up")
		facing_temp = "right";
	else if (facing_temp == "right")
		facing_temp = "down";
	return facing_temp;
}

std::string AverageAgent::turnLeft(std::string facing_temp){
	if (facing_temp == "down")
		facing_temp = "right";
	else if (facing_temp == "left")
		facing_temp = "down";
	else if (facing_temp == "up")
		facing_temp = "left";
	else if (facing_temp == "right")
		facing_temp = "up";
	return facing_temp;
}

Agent::Action AverageAgent::getAction(){
	if (checkPerceptIn(Agent::BUMP)){
		updateBoardSize();
	}
	placeExaminePercepts();

	// Climb out of the cave if agent has gold and is @ starting cell
	//	OR if there is BREEZE @ starting cell
	if ((position.first == 0 && position.second == 0) && (checkPerceptIn(Agent::BREEZE) || gold)){
		return CLIMB;
	}

	if (checkPerceptIn(Agent::SCREAM)){
		wumpusKilled = true;
		long_way_home.push_back(FORWARD);
		return FORWARD;
	}

	move_num++;
	if (!move_proc.empty()){
		Action a = move_proc[0];
		move_proc.erase(move_proc.begin());
		long_way_home.push_back(a);
		return a;
	}

	if (gold){
		if (position.first == 0 && position.second == 0){
			return CLIMB;
		}
		// Backtrack
		move_proc = long_way_home;
		move_proc.push_back(TURN_RIGHT);
		std::reverse(move_proc.begin(), move_proc.end());

		for (int i = 0; i < move_proc.size(); i++){
			if (move_proc[i] == TURN_LEFT){
				move_proc[i] = TURN_RIGHT;
			}
			else if (move_proc[i] == TURN_RIGHT){
				move_proc[i] = TURN_LEFT;
			}
		}
		return TURN_LEFT;
	}

	if (move_num == 1){
		if (checkPerceptIn(Agent::STENCH) && notShot && !wumpusKilled){
			//std::cout << "FIRE!" << std::endl;
			return SHOOT;
		}
		long_way_home.push_back(FORWARD);
		return FORWARD;
	}

	if (checkPerceptIn(Agent::GLITTER)){
		gotGold = true;
		gold = true;
		return PICK_UP;
	}

	std::pair<std::string, Cell> d = decideMove();
	//std::cout << "Decided to move" << d.second.x << d.second.y << std::endl;
	planMove(d);
	Action move = move_proc[0];
	move_proc.erase(move_proc.begin());
	long_way_home.push_back(move);
	return move;
}

void AverageAgent::perceiveResult(std::set<Percept> percepts){
	ps = percepts;
}