/* Author: Minjae Wee	mwee@uci.edu */
// Wumpus World - Smart AI.cpp

#include "SmartAgent.hpp"

SmartAgent::SmartAgent()
	:Agent()
{
	move_num = 0;
	row = 25;
	col = 25;
	for (int i = 0; i < row; i++){
		std::vector<Cell> temp;
		for (int j = 0; j < col; j++){
			Cell c;
			c.x = i;
			c.y = j;
			c.pit = 0;
			c.wumpus = 0;
			c.stench = -1;
			c.breeze = -1;
			c.visited = 0;
			temp.push_back(c);
		}
		board.push_back(temp);
	}
	board[0][0].stench = 0;
	board[0][0].breeze = 0;
	board[0][0].visited = 1;
	position = std::pair<int, int>(0, 0);	// start @ cell (0,0)
	facing = std::pair<int, int>(0, 1);		// start facing right
	saved_coord = std::pair<int, int>(-1, -1);

	gold = false;
	gotGold = false;
	notShot = true;
	wumpusKilled = false;
	certain_wumpus = false;
	saved_num = -1;
	lastAction = NIL;
}

// Returns true if the given pair is inside a vector of pairs; otherwise false
bool SmartAgent::checkCoordIn(std::pair<int, int> coord, std::vector<std::pair<int, int>> v){
	return (std::find(v.begin(), v.end(), coord) != v.end());
}

// Returns true if a given percept is @ current position; otherwise false
bool SmartAgent::checkPerceptIn(Percept p){
	const bool is_in = ps.find(p) != ps.end();
	return is_in;
}

// Called when agent visits a cell
void SmartAgent::visitCell(){
	for (int i = 0; i < row; i++){
		for (int j = 0; j < col; j++){
			board[i][j].wumpus = 0;
			board[i][j].pit = 0;
		}
	}

	for (int i = 0; i < row; i++){
		for (int j = 0; j < col; j++){
			if (board[i][j].stench == 1){
				//std::cout << "updated wumpus likelihood" << std::endl;
				updateWumpusPitProb(i, j, 'W', 1, false);
			}
			if (board[i][j].breeze == 1){
				//std::cout << "updated pit likelihood" << std::endl;
				updateWumpusPitProb(i, j, 'P', 1, false);
			}
		}
	}

	for (int i = 0; i < row; i++){
		for (int j = 0; j < col; j++){
			// Visited already => safe (no pit, no wumpus)
			if (board[i][j].visited){
				board[i][j].wumpus = 0;
				board[i][j].pit = 0;
				if (board[i][j].stench == 0)
					updateWumpusPitProb(i, j, 'W', 0, true);
				if (board[i][j].breeze == 0)
					updateWumpusPitProb(i, j, 'P', 0, true);
			}
		}
	}

	for (int i = 0; i < row; i++){
		for (int j = 0; j < col; j++){
			if (board[i][j].visited){
				board[i][j].wumpus = 0;
				board[i][j].pit = 0;
			}
		}
	}
}

// Updates the likelihood of wumpus and pit at a particular cell (row, column)
void SmartAgent::updateWumpusPitProb(int r, int c, char wum_or_pit, int prob, bool setvalue){
	std::vector<std::pair<int, int>> temp_adj_cells;
	temp_adj_cells.push_back(std::pair<int, int>(r, c - 1));
	temp_adj_cells.push_back(std::pair<int, int>(r, c + 1));
	temp_adj_cells.push_back(std::pair<int, int>(r - 1, c));
	temp_adj_cells.push_back(std::pair<int, int>(r + 1, c));

	// Check if adjacent cell coordinates are on board
	std::vector<std::pair<int, int>> adj_cells;
	for (int i = 0; i < temp_adj_cells.size(); i++){
		int a = temp_adj_cells[i].first;
		int b = temp_adj_cells[i].second;
		if (a >= 0 && a < row && b >= 0 && b < col){
			adj_cells.push_back(std::pair<int, int>(a, b));
		}
	}
		
	for (int i = 0; i < adj_cells.size(); i++){
		int a = adj_cells[i].first;
		int b = adj_cells[i].second;
		if (wum_or_pit == 'W' && board[a][b].wumpus != 4){ ////////////////////////
			if (!setvalue)
				board[a][b].wumpus += prob;
			else
				board[a][b].wumpus = prob;
		}
		else if (wum_or_pit == 'P' && board[a][b].pit != 4){ ///////////////////////
			if (!setvalue)
				board[a][b].pit += prob;
			else
				board[a][b].pit = prob;
		}
	}
}

// Updates the BREEZE and STENCH location
void SmartAgent::placeExaminePercepts(){
	if (checkPerceptIn(Agent::STENCH)){
		board[position.first][position.second].stench = 1;
	}
	else{
		board[position.first][position.second].stench = 0;
	}
	if (checkPerceptIn(Agent::BREEZE)){
		board[position.first][position.second].breeze = 1;
	}
	else{
		board[position.first][position.second].breeze = 0;
	}
	visitCell();
	if (move_proc.empty()){
		board[position.first][position.second].visited++;
	}
}

// Resizes the board - called when agent senses a BUMP
void SmartAgent::updateBoardSize(){
	// if agent facing right
	if (facing.first == 0 && facing.second == 1){
		col = position.second + 1;
		//std::cout << "resized/updated column to: " << col << std::endl;
		for (int j = 0; j < board.size(); j++)
		{
			board[j].resize(col);
		}
	}
	// if agent facing down
	if (facing.first == 1 && facing.second == 0){
		row = position.first + 1;
		//std::cout << "resized/updated row to: " << row << std::endl;
		board.resize(row);
	}
}

// Return a list of actions that need to be taken to travel to the given cell
void SmartAgent::planMove(std::pair<std::string, Cell> move){
	//std::cout << "\t\tPlanning to move: " << move.first << move.second.x << move.second.y << std::endl;
	if (move.first == "climb"){
		move_proc.clear();
		move_proc.push_back(CLIMB);
		return;
	}

	if (move.first == "pass"){
		move_proc.clear();
		move_proc.push_back(TURN_RIGHT);
		return;
	}

	std::string goal = facingToString(facing);
	if (move.second.x < position.first){
		goal = "up";
	}
	if (move.second.x > position.first){
		goal = "down";
	}
	if (move.second.y < position.second){
		goal = "left";
	}
	if (move.second.y > position.second){
		goal = "right";
	}

	std::vector<Action> temp_proc1;
	std::vector<Action> temp_proc2;
	std::string temp_facing;

	temp_facing = facingToString(facing);
	while (temp_facing != goal){
		//std::cout << "\t\t\tGoal = " << goal << "\t\tTempFacing = " << temp_facing << std::endl;
		temp_proc1.push_back(TURN_RIGHT);
		temp_facing = turnRight(temp_facing);
	}

	temp_facing = facingToString(facing);
	while (temp_facing != goal){
		temp_proc2.push_back(TURN_LEFT);
		temp_facing = turnLeft(temp_facing);
	}

	if (temp_proc1.size() < temp_proc2.size())
		move_proc = temp_proc1;
	else
		move_proc = temp_proc2;

	if (move.first == "shoot")
		move_proc.push_back(SHOOT);
	else if (move.first == "climb")
		move_proc.push_back(CLIMB);
	else
		move_proc.push_back(FORWARD);
}

// Depth First Search Traversal from GOLD to starting cell
void SmartAgent::dft(std::vector<std::pair<int, int>> safe_cells, std::pair<int, int> start){
	shortcut.push_back(start);
	for (int i = 0; i < safe_cells.size(); i++){
		visited[safe_cells[i]] = false;
	}
	std::pair<int, int> curr = start;
	visited[curr] = true;
	std::vector<std::pair<int, int>> n = getAdjCells(curr, safe_cells);
	for (int i = 0; i < n.size(); i++){
		if (!visited[n[i]]){
			DFTr(safe_cells, n[i]);
		}
	}
}

// Recursive Depth First Search
void SmartAgent::DFTr(std::vector<std::pair<int, int>> safe_cells, std::pair<int, int> v){
	shortcut.push_back(v);
	visited[v] = true;
	std::vector<std::pair<int, int>> n;
	n = getAdjCells(v, safe_cells);
	for (int i = 0; i < n.size(); i++){
		if (!visited[n[i]]){
			DFTr(safe_cells, n[i]);
		}
	}
}

// Determines the ideal ("smart") move and cell to visit
std::pair<std::string, Cell> SmartAgent::decideMove(){
	Cell none;
	if (gold && position.first == 0 && position.second == 0){
		return std::pair<std::string, Cell>("climb", none);
	}

	std::vector<Cell> possible_moves = getPossibleMoves();
	std::sort(possible_moves.begin(), possible_moves.end(), sort_danger());

	for (int i = 0; i < possible_moves.size(); i++){
		if (possible_moves[i].wumpus >= 2 && !wumpusKilled && notShot){
			saved_num = move_num;
			saved_coord.first = position.first;
			saved_coord.second = position.second;
			return std::pair<std::string, Cell>("shoot", possible_moves[i]);
		}
	}

	for (int i = 0; i < possible_moves.size(); i++){
		if (!possible_moves[i].visited && !possible_moves[i].wumpus && !possible_moves[i].pit){
			if (possible_moves[i].x == wumpus_pos.first && possible_moves[i].y == wumpus_pos.second)
				continue;
			return std::pair<std::string, Cell>("move", possible_moves[i]);
		}
	}

	for (int i = 0; i < possible_moves.size(); i++){
		if (possible_moves[i].visited <= 1 && possible_moves[i].wumpus <= 1 && possible_moves[i].pit == 0){
			if (possible_moves[i].wumpus != 0 && notShot){
				saved_num = move_num; 
				saved_coord.first = position.first;
				saved_coord.second = position.second;
				return std::pair<std::string, Cell>("shoot", possible_moves[i]);
			}
			if (possible_moves[i].x == wumpus_pos.first && possible_moves[i].y == wumpus_pos.second)
				continue;
			return std::pair<std::string, Cell>("move", possible_moves[i]);
		}
	}

	for (int i = 0; i < possible_moves.size(); i++){
		if (possible_moves[i].visited <= 2 && possible_moves[i].wumpus <= 2 && possible_moves[i].pit == 0){
			if (possible_moves[i].wumpus != 0 && notShot){
				saved_num = move_num;
				saved_coord.first = position.first;
				saved_coord.second = position.second;
				return std::pair<std::string, Cell>("shoot", possible_moves[i]);
			}
			if (possible_moves[i].x == wumpus_pos.first && possible_moves[i].y == wumpus_pos.second) 
				continue;
			return std::pair<std::string, Cell>("move", possible_moves[i]);
		}
	}

	for (int i = 0; i < possible_moves.size(); i++){
		if (possible_moves[i].visited <= 3 && possible_moves[i].wumpus <= 3 && possible_moves[i].pit == 0){
			if (possible_moves[i].wumpus != 0 && notShot){
				saved_num = move_num;
				saved_coord.first = position.first;
				saved_coord.second = position.second;
				return std::pair<std::string, Cell>("shoot", possible_moves[i]);
			}
			if (possible_moves[i].x == wumpus_pos.first && possible_moves[i].y == wumpus_pos.second)
				continue;
			return std::pair<std::string, Cell>("move", possible_moves[i]);
		}
	}
	// std::cout << "No safe move to make. Start backtracking!" << std::endl;
	gold = true;
	return std::pair<std::string, Cell>("pass", none);
}



// Optimizes the shortcut path outputted by DFS traversal and returns safe shortest path from GOLD to starting cell
std::vector<std::pair<int,int>> SmartAgent::optimizeShortcut(){

	std::vector<std::pair<int, int>> temp;
	for (int i = 0; i < shortcut.size(); i++){
		temp.push_back(shortcut[i]);
		if (shortcut[i].first == 0 && shortcut[i].second == 0){
			break;
		}
	}
	shortcut = temp;

	/*
	std::cout << "\n\n\nmid-shortcut" << std::endl;
	for (int i = 0; i < shortcut.size(); i++){
		std::cout << "\t(" << shortcut[i].first << "," << shortcut[i].second << ")" << std::endl;
	}
	*/

	std::vector<std::vector<int>> result;
	std::vector<int> index_to_pop;
	std::pair<int, int> prev;
	std::pair<int, int> after;
	int px, py, ax, ay;
	for (int i = 0; i < shortcut.size(); i++){
		prev = shortcut[i];
		for (int j = i+2; j < shortcut.size(); j++){
			after = shortcut[j];
			px = prev.first;
			py = prev.second;
			ax = after.first;
			ay = after.second;
			if ((px == ax && abs(py - ay) == 1) || (abs(px - ax) == 1 && py == ay)){
				for (int k = i + 1; k < j; k++){
					index_to_pop.push_back(k);
				}
				if (index_to_pop.size() > result.size()){
					result.clear();
					result.push_back(index_to_pop);
				}
				if (index_to_pop.size() == result.size()){
					result.push_back(index_to_pop);
				}
				index_to_pop.clear();
			}
		}
	}
	if (result.empty()){
		shortcut.erase(shortcut.begin());
		return shortcut;
	}

	int z;
	if (result.size() == 1){
		//std::cout << "\n\n\nindex_to_pop" << std::endl;
		z = 0;
		for (int i = 0; i < result[0].size(); i++){
			std::cout << "Attempting to pop: " << result[0][i] << std::endl;
			shortcut.erase(shortcut.begin() + (result[0][i] - z)); // <<<<<<<<<
			z++;
		}
		shortcut.erase(shortcut.begin());
		return shortcut;	
	}

	else if (result.size() > 1){
		std::vector<std::pair<int, int>> temp_shortcut;
		std::vector<std::vector<std::pair<int, int>>> potential_shortcut;
		//std::cout << "\n\n\nindex_to_pop" << std::endl;
		z = 0;
		for (int j = 0; j < result.size(); j++){
			std::vector<std::pair<int, int>> temp_shortcut = shortcut;
			for (int i = 0; i < result[j].size(); i++){
				std::cout << "Attempting to pop: " << result[j][i] << std::endl;
				temp_shortcut.erase(temp_shortcut.begin() + (result[j][i] - z)); // <<<<<<<<<
				z++;
			}
			potential_shortcut.push_back(temp_shortcut);
		}

		int best;
		for (int i = 0; i < potential_shortcut.size(); i++){
			for (int j = 0; j < potential_shortcut[i].size() - 1; j++){
				px = potential_shortcut[i][j].first;
				py = potential_shortcut[i][j].second;
				ax = potential_shortcut[i][j + 1].first;
				ay = potential_shortcut[i][j + 1].second;
				if (!((px == ax && abs(py - ay) == 1) || (abs(px - ax) == 1 && py == ay))){
					break;
				}
				else{
					best = i;
				}
			}
		}
		/*
		for (int i = 0; i < potential_shortcut[best].size(); i++){
			std::cout << "(" << potential_shortcut[best][i].first << "," << potential_shortcut[best][i].second << ")" << std::endl;
		}
		*/
		shortcut = potential_shortcut[best];
		shortcut.erase(shortcut.begin());
		return shortcut;
	}
}

// Return all 2 ~ 4 adjacent cell positions, given the current cell position
std::vector<Cell> SmartAgent::getPossibleMoves(){
	std::vector<Cell> cells;
	int x = position.first;
	int y = position.second;
	if (x >= 0 && x < row && y - 1 >= 0 && y - 1 < col){ // left cell
		cells.push_back(board[x][y - 1]);
	}
	if (x >= 0 && x < row && y + 1 >= 0 && y + 1 < col){ // right cell 
		cells.push_back(board[x][y + 1]);
	}
	if (x - 1 >= 0 && x - 1 < row && y >= 0 && y < col){ // top cell 
		cells.push_back(board[x - 1][y]);
	}
	if (x + 1 >= 0 && x + 1 < row && y >= 0 && y < col){ // bottom cell
		cells.push_back(board[x + 1][y]);

	}
	/*
	for (int i = 0; i < cells.size(); i++){
		std::cout << "Possible move" << i << ": " << cells[i].x << cells[i].y << "\n\n";
	}
	*/
	return cells;
}

// Return adjacent cells to a given cell (used to calculate the shortest path)
std::vector<std::pair<int, int>> SmartAgent::getAdjCells(std::pair<int, int> pos, std::vector<std::pair<int, int>> safe){
	std::vector<std::pair<int, int>> result;
	int x = pos.first;
	int y = pos.second;
	std::pair<int, int> c;
	if (x >= 0 && x < row && y - 1 >= 0 && y - 1< col){ // top cell 
		c.first = x;
		c.second = y-1;
		if (checkCoordIn(c, safe)){
			result.push_back(c);
		}
	}
	if (x - 1 >= 0 && x - 1 < row && y >= 0 && y < col){ // left cell
		c.first = x-1;
		c.second = y;
		if (checkCoordIn(c, safe)){
			result.push_back(c);
		}
	}
	if (x + 1 >= 0 && x + 1 < row && y >= 0 && y < col){ // right cell 
		c.first = x + 1;
		c.second = y;
		if (checkCoordIn(c, safe)){
			result.push_back(c);
		}
	}
	if (x >= 0 && x < row && y + 1 >= 0 && y + 1< col){ // bottom cell
		c.first = x;
		c.second = y + 1;
		if (checkCoordIn(c, safe)){
			result.push_back(c);
		}
	}
	return result;
}

// Returns a vector of Actions that need to be taken to get to the destination
std::vector<Agent::Action> SmartAgent::actionsToTake(std::pair<int, int> start, std::pair<int, int> dest, std::pair<int, int> dir){
	int x, y;
	std::string goal;
	std::vector<Action> result;
	std::vector<Action> a1;
	std::vector<Action> a2;
	std::string curr = facingToString(dir);
	if (dest.first - start.first == 1){
		goal = "down";
	}
	else if (dest.first - start.first == -1){
		goal = "up";
	}
	else if (dest.second - start.second == 1){
		goal = "right";
	}
	else if (dest.second - start.second == -1){
		goal = "left";
	}

	while (curr != goal){
		a1.push_back(TURN_RIGHT);
		curr = turnRight(curr);
	}

	curr = facingToString(dir);
	while (curr != goal){
		a2.push_back(TURN_LEFT);
		curr = turnLeft(curr);
	}

	if (a1.size() < a2.size()){
		result = a1;
	}
	else{
		result = a2;
	}
	if (curr == goal){
		result.push_back(FORWARD);
		return result;
	}
	return result;
}

// Converts facing pair into string format
std::string SmartAgent::facingToString(std::pair<int, int> facing){
	std::string result;
	if (facing.first == 1 && facing.second == 0)
		result = "down";
	else if (facing.first == -1 && facing.second == 0)
		result = "up";
	else if (facing.first == 0 && facing.second == 1)
		result = "right";
	else if (facing.first == 0 && facing.second == -1)
		result = "left";
	return result;
}

// Agent turns right 90 degrees
std::string SmartAgent::turnRight(std::string facing_temp){
	if (facing_temp == "down")
		facing_temp = "left";
	else if (facing_temp == "left")
		facing_temp = "up";
	else if (facing_temp == "up")
		facing_temp = "right";
	else if (facing_temp == "right")
		facing_temp = "down";
	return facing_temp;
}

// Agent turns left 90 degrees
std::string SmartAgent::turnLeft(std::string facing_temp){
	if (facing_temp == "down")
		facing_temp = "right";
	else if (facing_temp == "left")
		facing_temp = "down";
	else if (facing_temp == "up")
		facing_temp = "left";
	else if (facing_temp == "right")
		facing_temp = "up";
	return facing_temp;
}

// Returns a single action that the agent will take
Agent::Action SmartAgent::getAction(){
	if (checkPerceptIn(Agent::BUMP)){
		updateBoardSize();
	}
	placeExaminePercepts();

	// Climb out of the cave if agent has gold and is @ starting cell
	//	OR if there is BREEZE @ starting cell
	if ((position.first == 0 && position.second == 0) && (checkPerceptIn(Agent::BREEZE) || gold)){
		return CLIMB;
	}

	move_num++;
	if (!move_proc.empty()){
		Action a = move_proc[0];
		move_proc.erase(move_proc.begin());
		return a;
	}

	if (gold){
		if (position.first == 0 && position.second == 0){
			return CLIMB;
		}
		std::vector<std::pair<int, int>> safe_cells;
		for (int i = 0; i < row; i++){
			for (int j = 0; j < col; j++){
				if (board[i][j].visited){
					safe_cells.push_back(std::pair<int, int>(i, j));
				}
			}
		}

		dft(safe_cells, position);
		
		/*
		std::cout << "\n\n\nInefficient shortcut" << std::endl;
		for (int i = 0; i < shortcut.size(); i++){
			std::cout << "\t(" << shortcut[i].first << "," << shortcut[i].second << ")" << std::endl;
		}
		*/

		optimizeShortcut();
		
		/*
		std::cout << "\n\n\nOptimized shortcut" << std::endl;
		for (int i = 0; i < shortcut.size(); i++){
			std::cout << "\t(" << shortcut[i].first << "," << shortcut[i].second << ")" << std::endl;
		}
		*/
		
		

		if (!shortcut.empty()){
			std::vector<Action> moves = actionsToTake(position, shortcut[0], facing);
			move_proc = moves;
			Action move = move_proc[0];
			move_proc.erase(move_proc.begin());
			return move;
		}
	}

	if (move_num == 1){
		if (checkPerceptIn(Agent::STENCH) && notShot && !wumpusKilled){
			saved_num = move_num;/////////
			saved_coord.first = position.first;
			saved_coord.second = position.second;
			return SHOOT;
		}
		return FORWARD;
	}

	if (!certain_wumpus && move_num == saved_num + 1){
		if (!notShot && !wumpusKilled && saved_coord.first >= 0 && saved_coord.first < row &&  saved_coord.second >= 0 && saved_coord.second < col){
			//std::cout << "WUMPUS IS PRESENT AT" << saved_coord.first + 1 << "," << saved_coord.second << std::endl;
			wumpus_pos = std::pair<int,int>(saved_coord.first+1, saved_coord.second);
			saved_num = -1;
			certain_wumpus = true;
		}
	}

	if (checkPerceptIn(Agent::GLITTER)){
		gotGold = true;
		gold = true;
		return PICK_UP;
	}

	std::pair<std::string, Cell> d = decideMove();
	//std::cout << "Decided to move" << d.second.x << d.second.y << std::endl;
	planMove(d);
	Action move = move_proc[0];
	move_proc.erase(move_proc.begin());
	return move;
}

// Insert & update all the percepts at current cell into agent's percepts
void SmartAgent::perceiveResult(std::set<Percept> percepts){
	ps = percepts;
}